"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tsutils_1 = require("tsutils");
const typescript_1 = require("typescript");
const unsafe_assignment_rule_1 = require("./unsafe-assignment-rule");
// eslint-disable-next-line functional/functional-parameters
const unsafeIndexAssignmentFunc = () => false;
const unsafePropertyAssignmentFunc = (
// eslint-disable-next-line @typescript-eslint/ban-types
destinationProperty, 
// eslint-disable-next-line @typescript-eslint/ban-types
sourceProperty) => {
    const destinationPropIsOptional = tsutils_1.isSymbolFlagSet(destinationProperty, typescript_1.SymbolFlags.Optional);
    return destinationPropIsOptional && sourceProperty === undefined;
};
const message = "Assigning to an optional property when there is no such property in the source type (optional or otherwise) is unsafe.";
/**
 * An ESLint rule to ban unsafe assignment to optional properties.
 */
const noUnsafeOptionalPropertyAssignment = {
    meta: {
        type: "problem",
        docs: {
            category: "Possible Errors",
            description: "Bans unsafe assignment to optional properties.",
            recommended: "error",
            url: "https://github.com/danielnixon/eslint-plugin-total-functions",
        },
        messages: {
            errorStringCallExpression: message,
            errorStringAssignmentExpression: message,
            errorStringVariableDeclaration: message,
            errorStringArrowFunctionExpression: message,
            errorStringTSAsExpression: message,
            errorStringTSTypeAssertion: message,
        },
        schema: [],
    },
    create: unsafe_assignment_rule_1.createNoUnsafeAssignmentRule(unsafePropertyAssignmentFunc, unsafeIndexAssignmentFunc),
};
exports.default = noUnsafeOptionalPropertyAssignment;
